# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
import os

import pytest
from maggma.stores import JSONStore

from pymatgen.analysis.diffusion.utils.maggma import get_entries_from_dbs

dir_path = os.path.dirname(os.path.realpath(__file__)) + "/test_files"

__author__ = "Jimmy Shen"
__version__ = "1.0"
__date__ = "April 10, 2019"


@pytest.fixture
def maggma_stores():
    return {
        "sgroups": JSONStore(f"{dir_path}/maggma_sgroup_store.json", key="group_id"),
        "materials": JSONStore(f"{dir_path}/maggma_materials_store.json", key="material_id"),
    }


def test(maggma_stores):
    base_ents, inserted_ents = get_entries_from_dbs(
        maggma_stores["sgroups"], maggma_stores["materials"], "Mg", material_id="mvc-6910_Mg"
    )

    # check that the entries have been created
    def has_mg(ent):
        return "Mg" in ent.composition.as_dict().keys()

    assert all(map(has_mg, inserted_ents))
    assert not any(map(has_mg, base_ents))



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
import os
import unittest

import numpy as np
from monty.serialization import loadfn
from pymatgen.core import PeriodicSite, Structure
from pymatgen.io.vasp import Chgcar

from pymatgen.analysis.diffusion.neb.full_path_mapper import (
    ChargeBarrierGraph,
    MigrationGraph,
    MigrationHop,
    get_hop_site_sequence,
    order_path,
)
from pymatgen.analysis.diffusion.neb.periodic_dijkstra import _get_adjacency_with_images

dir_path = os.path.dirname(os.path.realpath(__file__))

__author__ = "Jimmy Shen"
__version__ = "1.0"
__date__ = "April 10, 2019"


class MigrationGraphSimpleTest(unittest.TestCase):
    def setUp(self):
        struct = Structure.from_file(f"{dir_path}/full_path_files/MnO2_full_Li.vasp")
        self.fpm = MigrationGraph.with_distance(structure=struct, migrating_specie="Li", max_distance=4)

    def test_get_pos_and_migration_hop(self):
        """
        Make sure that we can populate the graph with MigrationHop Objects
        """
        self.fpm._get_pos_and_migration_hop(0, 1, 1)
        self.assertAlmostEqual(self.fpm.m_graph.graph[0][1][1]["hop"].length, 3.571248, 4)

    def test_get_summary_dict(self):
        summary_dict = self.fpm.get_summary_dict()
        self.assertTrue("hop_label", summary_dict["hops"][0])
        self.assertTrue("hop_label", summary_dict["unique_hops"][0])


class MigrationGraphFromEntriesTest(unittest.TestCase):
    def setUp(self):
        self.test_ents_MOF = loadfn(f"{dir_path}/full_path_files/Mn6O5F7_cat_migration.json")
        self.aeccar_MOF = Chgcar.from_file(f"{dir_path}/full_path_files/AECCAR_Mn6O5F7.vasp")
        self.li_ent = loadfn(f"{dir_path}/full_path_files/li_ent.json")["li_ent"]
        entries = [self.test_ents_MOF["ent_base"]] + self.test_ents_MOF["one_cation"]
        self.full_struct = MigrationGraph.get_structure_from_entries(
            entries=entries,
            migrating_ion_entry=self.li_ent,
        )[0]

    def test_m_graph_from_entries_failed(self):
        # only base
        s_list = MigrationGraph.get_structure_from_entries(
            entries=[self.test_ents_MOF["ent_base"]],
            migrating_ion_entry=self.li_ent,
        )
        self.assertEqual(len(s_list), 0)

        s_list = MigrationGraph.get_structure_from_entries(
            entries=self.test_ents_MOF["one_cation"],
            migrating_ion_entry=self.li_ent,
        )
        self.assertEqual(len(s_list), 0)

    def test_m_graph_construction(self):
        self.assertEqual(self.full_struct.composition["Li"], 8)
        mg = MigrationGraph.with_distance(self.full_struct, migrating_specie="Li", max_distance=4.0)
        self.assertEqual(len(mg.m_graph.structure), 8)


class MigrationGraphComplexTest(unittest.TestCase):
    def setUp(self):
        struct = Structure.from_file(f"{dir_path}/full_path_files/MnO2_full_Li.vasp")
        self.fpm_li = MigrationGraph.with_distance(structure=struct, migrating_specie="Li", max_distance=4)

        # Particularity difficult pathfinding since both the starting and ending positions are outside the unit cell
        struct = Structure.from_file(f"{dir_path}/full_path_files/Mg_2atom.vasp")
        self.fpm_mg = MigrationGraph.with_distance(structure=struct, migrating_specie="Mg", max_distance=2)

    def test_group_and_label_hops(self):
        """
        Check that the set of end points in a group of similiarly labeled hops are all the same
        """
        edge_labs = np.array([d["hop_label"] for u, v, d in self.fpm_li.m_graph.graph.edges(data=True)])

        site_labs = np.array(
            [
                (
                    d["hop"].symm_structure.wyckoff_symbols[d["hop"].iindex],
                    d["hop"].symm_structure.wyckoff_symbols[d["hop"].eindex],
                )
                for u, v, d in self.fpm_li.m_graph.graph.edges(data=True)
            ]
        )

        for itr in range(edge_labs.max()):
            sub_set = site_labs[edge_labs == itr]
            for end_point_labels in sub_set:
                self.assertTrue(sorted(end_point_labels) == sorted(sub_set[0]))

    def test_unique_hops_dict(self):
        """
        Check that the unique hops are inequilvalent
        """
        unique_list = [v for k, v in self.fpm_li.unique_hops.items()]
        all_pairs = [(mg1, mg2) for i1, mg1 in enumerate(unique_list) for mg2 in unique_list[i1 + 1 :]]

        for migration_hop in all_pairs:
            self.assertNotEqual(migration_hop[0]["hop"], migration_hop[1]["hop"])

    def test_add_data_to_similar_edges(self):
        # passing normal data
        self.fpm_li.add_data_to_similar_edges(0, {"key0": "data"})
        for u, v, d in self.fpm_li.m_graph.graph.edges(data=True):
            if d["hop_label"] == 0:
                self.assertEqual(d["key0"], "data")

        # passing ordered list data
        migration_hop = self.fpm_li.unique_hops[1]["hop"]
        self.fpm_li.add_data_to_similar_edges(1, {"key1": [1, 2, 3]}, m_hop=migration_hop)
        for u, v, d in self.fpm_li.m_graph.graph.edges(data=True):
            if d["hop_label"] == 1:
                self.assertEqual(d["key1"], [1, 2, 3])

        # passing ordered list with direction
        migration_hop_reversed = MigrationHop(
            isite=migration_hop.esite,
            esite=migration_hop.isite,
            symm_structure=migration_hop.symm_structure,
        )
        self.fpm_li.add_data_to_similar_edges(2, {"key2": [1, 2, 3]}, m_hop=migration_hop_reversed)
        for u, v, d in self.fpm_li.m_graph.graph.edges(data=True):
            if d["hop_label"] == 2:
                self.assertEqual(d["key2"], [3, 2, 1])

    def test_assign_cost_to_graph(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'
        for u, v, d in self.fpm_li.m_graph.graph.edges(data=True):
            self.assertAlmostEqual(d["cost"], d["hop_distance"], 4)

        self.fpm_li.assign_cost_to_graph(cost_keys=["hop_distance", "hop_distance"])
        for u, v, d in self.fpm_li.m_graph.graph.edges(data=True):
            self.assertAlmostEqual(d["cost"], d["hop_distance"] ** 2, 4)

    def test_periodic_dijkstra(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'

        # test the connection graph
        sgraph = self.fpm_li.m_graph
        G = sgraph.graph.to_undirected()
        conn_dict = _get_adjacency_with_images(G)
        for u in conn_dict.keys():
            for v in conn_dict[u]:
                for k, d in conn_dict[u][v].items():
                    neg_image = tuple(-dim_ for dim_ in d["to_jimage"])
                    opposite_connections = [d2_["to_jimage"] for k2_, d2_ in conn_dict[v][u].items()]
                    self.assertIn(neg_image, opposite_connections)

    def test_get_path(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'
        paths = [*self.fpm_li.get_path(flip_hops=False)]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}
        self.assertIn("5->7->5", p_strings)
        # convert each pathway to a string representation
        paths = [*self.fpm_li.get_path(max_val=2.0, flip_hops=False)]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}

        # After checking trimming the graph more hops are needed for the same path
        self.assertIn("5->3->7->2->5", p_strings)

        self.fpm_mg.assign_cost_to_graph()  # use 'hop_distance'
        paths = [*self.fpm_mg.get_path(flip_hops=False)]
        p_strings = {"->".join(map(str, get_hop_site_sequence(ipath, start_u=u))) for u, ipath in paths}
        self.assertIn("1->0->1", p_strings)

    def test_get_key_in_path(self):
        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'
        paths = [*self.fpm_li.get_path(flip_hops=False)]
        hop_seq_info = [get_hop_site_sequence(ipath, start_u=u, key="hop_distance") for u, ipath in paths]

        hop_distances = {}
        for u, ipath in paths:
            hop_distances[u] = []
            for hop in ipath:
                hop_distances[u].append(hop["hop_distance"])

        # Check that the right key and respective values were pulled
        for u, distances in hop_distances.items():
            self.assertEqual(distances, hop_seq_info[u][1])
            print(distances, hop_seq_info[u][1])

    def test_not_matching_first(self):
        structure = Structure.from_file(f"{dir_path}/pathfinder_files/Li6MnO4.json")
        fpm_lmo = MigrationGraph.with_distance(structure, "Li", max_distance=4)
        for u, v, d in fpm_lmo.m_graph.graph.edges(data=True):
            self.assertIn(d["hop"].eindex, {0, 1})

    def test_order_path(self):
        # add list data to migration graph - to test if list data is flipped
        for n, hop_d in self.fpm_li.unique_hops.items():
            data = {"data": [hop_d["iindex"], hop_d["eindex"]]}
            self.fpm_li.add_data_to_similar_edges(n, data, hop_d["hop"])

        self.fpm_li.assign_cost_to_graph()  # use 'hop_distance'

        for n, hop_list in [*self.fpm_li.get_path(flip_hops=False)]:
            ordered = order_path(hop_list, n)

            # check if isites and esites are oriented to form a coherent path
            for h1, h2 in zip(ordered[:-1], ordered[1:]):
                self.assertEqual(h1["eindex"], h2["iindex"])

            # if hop was flipped, check list data was flipped too
            for h, ord_h in zip(hop_list, ordered):
                if h["iindex"] != ord_h["iindex"]:  # check only if hop was flipped
                    self.assertEqual(h["data"][0], ord_h["data"][-1])
                    self.assertEqual(h["data"][-1], ord_h["data"][0])


class ChargeBarrierGraphTest(unittest.TestCase):
    def setUp(self):
        self.full_sites_MOF = loadfn(f"{dir_path}/full_path_files/LixMn6O5F7_full_sites.json")
        self.aeccar_MOF = Chgcar.from_file(f"{dir_path}/full_path_files/AECCAR_Mn6O5F7.vasp")
        self.cbg = ChargeBarrierGraph.with_distance(
            structure=self.full_sites_MOF,
            migrating_specie="Li",
            max_distance=4,
            potential_field=self.aeccar_MOF,
            potential_data_key="total",
        )
        self.cbg._tube_radius = 10000

    def test_integration(self):
        """
        Sanity check: for a long enough diagonaly hop, if we turn the radius of the tube way up, it should cover the entire unit cell
        """
        total_chg_per_vol = (
            self.cbg.potential_field.data["total"].sum()
            / self.cbg.potential_field.ngridpts
            / self.cbg.potential_field.structure.volume
        )
        self.assertAlmostEqual(
            self.cbg._get_chg_between_sites_tube(self.cbg.unique_hops[2]["hop"]),
            total_chg_per_vol,
        )

        self.cbg._tube_radius = 2
        # self.cbg.populate_edges_with_chg_density_info()

        # find this particular hop
        ipos = [0.33079153, 0.18064031, 0.67945924]
        epos = [0.33587514, -0.3461259, 1.15269302]
        isite = PeriodicSite("Li", ipos, self.cbg.structure.lattice)
        esite = PeriodicSite("Li", epos, self.cbg.structure.lattice)
        ref_hop = MigrationHop(isite=isite, esite=esite, symm_structure=self.cbg.symm_structure)
        hop_idx = -1
        for k, d in self.cbg.unique_hops.items():
            if d["hop"] == ref_hop:
                hop_idx = k

        self.assertAlmostEqual(
            self.cbg._get_chg_between_sites_tube(self.cbg.unique_hops[hop_idx]["hop"]),
            0.188952739835188,
            3,
        )

    def test_populate_edges_with_chg_density_info(self):
        """
        Test that all of the sites with similar lengths have similar charge densities,
        this will not always be true, but it valid in this Mn6O5F7
        """
        self.cbg.populate_edges_with_chg_density_info()
        length_vs_chg = list(
            sorted((d["hop"].length, d["chg_total"]) for u, v, d in self.cbg.m_graph.graph.edges(data=True))
        )
        prv = None
        for length, chg in length_vs_chg:
            if prv is None:
                prv = (length, chg)
                continue

            if 1.05 > length / prv[0] > 0.95:
                self.assertAlmostEqual(chg, prv[1], 3)

    def test_get_summary_dict(self):
        summary_dict = self.cbg.get_summary_dict()
        self.assertTrue("chg_total", summary_dict["hops"][0])
        self.assertTrue("chg_total", summary_dict["unique_hops"][0])


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
"""
Functions for querying Materials Project style MongoStores that contains
cathode materials The functions are isolated from the rest of the package so
that the rest of the package will not depend on Maggma
"""

__author__ = "Jimmy Shen"
__copyright__ = "Copyright 2019, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Jimmy Shen"
__email__ = "jmmshn@lbl.gov"
__date__ = "July 21, 2019"

import logging
from typing import Union

from maggma.stores import MongoStore
from monty.serialization import MontyDecoder
from pymatgen.entries.computed_entries import ComputedEntry

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_entries_from_dbs(
    structure_group_store: MongoStore, material_store: MongoStore, migrating_ion: str, material_id: Union[str, int]
):
    """
    Get the entries needed to construct a migration from a database that
    contains topotactically matched structures.

    Args:
        structure_group_store: Electrode documents one per each similar group of
            insertion materials, can also use any db that contains a
        material_store: Material documenets one per each similar structure (
            multiple tasks)
        migrating_ion: The name of the migrating ion
        material_ids list with topotactic structures
    """

    with structure_group_store as store:
        sg_doc = store.query_one({structure_group_store.key: material_id})
    ignored_species = migrating_ion
    base_entries = []
    inserted_entries = []
    with material_store as store:
        for m_doc in store.query({"material_id": {"$in": sg_doc["material_ids"]}}):
            if "GGA+U" in m_doc["entries"]:
                entry = MontyDecoder().process_decoded(m_doc["entries"]["GGA+U"])  # type: ComputedEntry
            elif "GGA" in m_doc["entries"]:
                entry = MontyDecoder().process_decoded(m_doc["entries"]["GGA"])
            else:
                raise RuntimeError("Missing GGA or GGA+U calc type in <entries>")

            if ignored_species in entry.composition.as_dict().keys():
                inserted_entries.append(entry)
            else:
                base_entries.append(entry)

    return base_entries, inserted_entries



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

import os

import numpy as np
import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core.structure import PeriodicSite, Structure

from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
from pymatgen.analysis.diffusion.utils.edge_data_from_sc import add_edge_data_from_sc, get_uc_pos, get_unique_hop

test_dir = os.path.dirname(os.path.realpath(__file__))

__author__ = "Haoming Li"
__version__ = "1.0"
__date__ = "February 17, 2021"

uc_full_sites = Structure.from_file(f"{test_dir}/test_files/Li4Sr3Fe2O7_uc.vasp")
input_struct_i = Structure.from_file(f"{test_dir}/test_files/Sr3Fe2O7_sc_i.vasp")
input_struct_e = Structure.from_file(f"{test_dir}/test_files/Sr3Fe2O7_sc_e.vasp")

mg_uc_full_sites = Structure.from_file(f"{test_dir}/test_files/Mg3VOPO4_uc.vasp")
mg_input_struct_i = Structure.from_file(f"{test_dir}/test_files/Mg3VOPO4_sc_i.vasp")
mg_input_struct_e = Structure.from_file(f"{test_dir}/test_files/Mg3VOPO4_sc_e.vasp")

mg_Li = MigrationGraph.with_distance(structure=uc_full_sites, migrating_specie="Li", max_distance=5)
mg_Mg = MigrationGraph.with_distance(structure=mg_uc_full_sites, migrating_specie="Mg", max_distance=4)


def test_add_edge_data_from_sc():
    errors = []

    test_key = "test_key"
    test_array = [0, 1, 2, 3, 4]
    add_edge_data_from_sc(
        mg_Li,
        i_sc=input_struct_i,
        e_sc=input_struct_e,
        data_array=test_array,
        key=test_key,
    )

    edge_data = []
    for u, v, d in mg_Li.m_graph.graph.edges(data=True):
        edge_data.append(d)
    hop_labels = []
    for i in edge_data:
        for k, v in i.items():
            if k == test_key:
                hop_labels.append(i["hop_label"])

    if not hop_labels:
        errors.append("No data was added to any edge")
    if not all(i == hop_labels[0] for i in hop_labels):
        errors.append("Not all data are added to the same unique hop")

    assert not errors, "errors occured:\n" + "\n".join(errors)


def test_get_uc_pos():
    errors = []

    # set up parameters to initiate get_uc_pos
    uc_lattice = mg_Li.symm_structure.lattice
    isite = [x for x in input_struct_i.sites if x.species_string == "Li"][0]
    esite = [x for x in input_struct_e.sites if x.species_string == "Li"][0]
    sm = StructureMatcher(ignored_species=[list(mg_Li.m_graph.graph.edges(data=True))[0][2]["hop"].isite.specie.name])
    wi_specie = mg_Li.symm_structure[-1].specie

    p0, p1, p2 = get_uc_pos(isite, esite, mg_Li.symm_structure, input_struct_i, sm)

    # generate correct sites to compare
    test_p0 = PeriodicSite(
        wi_specie, np.array([2.91418875, 1.02974425, 4.4933425]), uc_lattice, coords_are_cartesian=True
    )
    test_p1 = PeriodicSite(
        wi_specie, np.array([4.82950555, 1.0247028, 4.10369437]), uc_lattice, coords_are_cartesian=True
    )
    test_p2 = PeriodicSite(
        wi_specie, np.array([6.74482475, 1.01967025, 3.7140425]), uc_lattice, coords_are_cartesian=True
    )

    if not test_p0.__eq__(p0):
        errors.append("Initial site does not match")
    if not test_p1.__eq__(p1):
        errors.append("Middle site does not match")
    if not test_p2.__eq__(p2):
        errors.append("Ending site does not match")

    assert not errors, "errors occured:\n" + "\n".join(errors)


def test_get_unique_hop_host():
    results = get_unique_hop(mg_Mg, mg_input_struct_i, mg_input_struct_i[0], mg_input_struct_e[0], use_host_sg=True)
    assert results[0] == 2


def test_get_unique_host_nonhost():
    with pytest.raises(Exception) as exc_info:
        get_unique_hop(mg_Mg, mg_input_struct_i, mg_input_struct_i[0], mg_input_struct_e[0], use_host_sg=False)
    expected_exception = "No symmetrically equivalent site was found for [0.53593472 2.8352428  4.54752366] Mg"
    assert exc_info.value.args[0] == expected_exception



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.


import itertools
import unittest

import numpy as np
from pymatgen.core import Lattice
from pymatgen.util.coord import pbc_diff
from scipy.cluster.vq import kmeans

from pymatgen.analysis.diffusion.aimd.clustering import Kmeans, KmeansPBC


class KmeansTest(unittest.TestCase):
    def test_cluster(self):
        data = np.random.uniform(size=(10, 5))
        data = list(data)
        d2 = np.random.uniform(size=(10, 5)) + ([5] * 5)
        data.extend(list(d2))
        d2 = np.random.uniform(size=(10, 5)) + ([-5] * 5)
        data.extend(list(d2))
        data = np.array(data)

        k = Kmeans()
        clusters = []
        for i in range(10):
            clusters.append(k.cluster(data, 3))
        c1, l1, ss = min(clusters, key=lambda d: d[2])
        c2, d = kmeans(data, 3)
        same = False
        for a in itertools.permutations(c2):
            if np.allclose(c1, a):
                same = True
                break
        self.assertTrue(same)


class KmeansPBCTest(unittest.TestCase):
    def test_cluster(self):
        lattice = Lattice.cubic(4)

        pts = []
        initial = [[0, 0, 0], [0.5, 0.5, 0.5], [0.25, 0.25, 0.25], [0.5, 0, 0]]
        for c in initial:
            for i in range(100):
                pts.append(np.array(c) + np.random.randn(3) * 0.01 + np.random.randint(3))
        pts = np.array(pts)
        k = KmeansPBC(lattice)
        centroids, labels, ss = k.cluster(pts, 4)
        for c1 in centroids:
            found = False
            for c2 in centroids:
                if np.allclose(pbc_diff(c1, c2), [0, 0, 0], atol=0.1):
                    found = True
                    break
            self.assertTrue(found)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.
import glob
import os
import unittest

import numpy as np
from pymatgen.core import Composition, PeriodicSite, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.testing import PymatgenTest

from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
from pymatgen.analysis.diffusion.neb.pathfinder import DistinctPathFinder, IDPPSolver, MigrationHop
from pymatgen.analysis.diffusion.utils.supercells import get_start_end_structures

__author__ = "Iek-Heng Chu"
__version__ = "1.0"
__date__ = "March 14, 2017"


def get_path(path_str, dirname="./"):
    cwd = os.path.abspath(os.path.dirname(__file__))
    path = os.path.join(cwd, dirname, path_str)
    return path


class IDPPSolverTest(unittest.TestCase):
    init_struct = Structure.from_file(get_path("CONTCAR-0", dirname="pathfinder_files"))
    final_struct = Structure.from_file(get_path("CONTCAR-1", dirname="pathfinder_files"))

    def test_idpp_from_ep(self):
        obj = IDPPSolver.from_endpoints([self.init_struct, self.final_struct], nimages=3, sort_tol=1.0)
        new_path = obj.run(
            maxiter=5000,
            tol=1e-5,
            gtol=1e-3,
            step_size=0.05,
            max_disp=0.05,
            spring_const=5.0,
            species=["Li"],
        )

        self.assertEqual(len(new_path), 5)
        self.assertEqual(new_path[1].num_sites, 111)
        self.assertTrue(
            np.allclose(
                new_path[0][2].frac_coords,
                np.array([0.50000014, 0.99999998, 0.74999964]),
            )
        )
        self.assertTrue(np.allclose(new_path[1][0].frac_coords, np.array([0.482439, 0.68264727, 0.26525066])))
        self.assertTrue(
            np.allclose(
                new_path[2][10].frac_coords,
                np.array([0.50113915, 0.74958704, 0.75147021]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[3][22].frac_coords,
                np.array([0.28422885, 0.62568764, 0.98975444]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[4][47].frac_coords,
                np.array([0.59767531, 0.12640952, 0.37745006]),
            )
        )

    def test_idpp_from_ep_diff_latt(self):

        # This is the same test as test_idpp_from_ep where we want to interpolate
        # different lattices.

        # make a copy of the final_struct and distort the lattice by 5%
        final_struct_strain = self.final_struct.copy()
        final_struct_strain.apply_strain(0.05)

        obj = IDPPSolver.from_endpoints(
            [self.init_struct, final_struct_strain],
            nimages=3,
            sort_tol=1.0,
            interpolate_lattices=True,
        )
        new_path = obj.run(
            maxiter=5000,
            tol=1e-5,
            gtol=1e-3,
            step_size=0.05,
            max_disp=0.05,
            spring_const=5.0,
            species=["Li"],
        )

        self.assertEqual(len(new_path), 5)
        self.assertEqual(new_path[1].num_sites, 111)

        # For checking coords, the endpoints will have the same frac coords as the
        # original test, whereas the midpoints will not have altered coord because
        # of the differing lattice.
        self.assertTrue(
            np.allclose(
                new_path[0][2].frac_coords,
                np.array([0.50000014, 0.99999998, 0.74999964]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[1][0].frac_coords,
                np.array([0.47483465, 0.6740367, 0.26270862]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[2][10].frac_coords,
                np.array([0.48839604, 0.73135841, 0.73315875]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[3][22].frac_coords,
                np.array([0.27395552, 0.60307242, 0.95398018]),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[4][47].frac_coords,
                np.array([0.59767531, 0.12640952, 0.37745006]),
            )
        )

        # ensure the lattices are interpolated
        self.assertTrue(
            np.allclose(
                new_path[0].lattice.matrix,
                np.array(
                    [
                        [10.4117233, 0.0, 0.0],
                        [0.0, 12.07835177, 0.0],
                        [0.0, 0.0, 9.47759962],
                    ]
                ),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[1].lattice.matrix,
                np.array(
                    [
                        [10.54186984, 0.0, 0.0],
                        [0.0, 12.22933117, 0.0],
                        [0.0, 0.0, 9.59606961],
                    ]
                ),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[2].lattice.matrix,
                np.array(
                    [
                        [10.67201638, 0.0, 0.0],
                        [0.0, 12.38031057, 0.0],
                        [0.0, 0.0, 9.71453961],
                    ]
                ),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[3].lattice.matrix,
                np.array(
                    [
                        [10.80216292, 0.0, 0.0],
                        [0.0, 12.53128996, 0.0],
                        [0.0, 0.0, 9.8330096],
                    ]
                ),
            )
        )
        self.assertTrue(
            np.allclose(
                new_path[4].lattice.matrix,
                np.array(
                    [
                        [10.93230946, 0.0, 0.0],
                        [0.0, 12.68226936, 0.0],
                        [0.0, 0.0, 9.9514796],
                    ]
                ),
            )
        )

    def test_idpp(self):
        images = self.init_struct.interpolate(self.final_struct, nimages=4, autosort_tol=1.0)
        obj = IDPPSolver(images)
        new_path = obj.run(
            maxiter=5000,
            tol=1e-5,
            gtol=1e-3,
            step_size=0.05,
            max_disp=0.05,
            spring_const=5.0,
            species=["Li"],
        )

        self.assertEqual(len(new_path), 5)
        self.assertEqual(new_path[1].num_sites, 111)
        self.assertTrue(
            np.allclose(
                new_path[0][2].frac_coords,
                np.array([0.50000014, 0.99999998, 0.74999964]),
            )
        )
        self.assertTrue(np.allclose(new_path[1][0].frac_coords, np.array([0.482439, 0.68264727, 0.26525066])))
        self.assertTrue(
            np.allclose(
                new_path[4][47].frac_coords,
                np.array([0.59767531, 0.12640952, 0.37745006]),
            )
        )


class DistinctPathFinderTest(PymatgenTest):
    def test_get_paths(self):
        s = self.get_structure("LiFePO4")
        # Only one path in LiFePO4 with 4 A.
        p = DistinctPathFinder(s, "Li", max_path_length=4)
        paths = p.get_paths()
        self.assertEqual(len(paths), 1)

        # Make sure this is robust to supercells.
        s.make_supercell((2, 2, 1))
        p = DistinctPathFinder(s, "Li", max_path_length=4)
        paths = p.get_paths()
        self.assertEqual(len(paths), 1)

        ss = paths[0].get_structures(vac_mode=False)
        self.assertEqual(len(ss), 7)

        paths[0].write_path("pathfindertest_noidpp_vac.cif", idpp=False)
        paths[0].write_path("pathfindertest_idpp_vac.cif", idpp=True)
        paths[0].write_path("pathfindertest_idpp_nonvac.cif", idpp=True, vac_mode=False)
        self.assertEqual(
            str(paths[0]),
            "Path of 3.0329 A from Li [0.000, 0.500, 1.000] (ind: 0, Wyckoff: 16a) to Li [-0.000, 0.750, 1.000] (ind: 0, Wyckoff: 16a)",
        )

        p = DistinctPathFinder(s, "Li", max_path_length=6)
        paths = p.get_paths()
        self.assertEqual(len(paths), 4)

        s = self.get_structure("Graphite")

        # Only one path in graphite with 2 A.
        p = DistinctPathFinder(s, "C0+", max_path_length=2)
        paths = p.get_paths()
        self.assertEqual(len(paths), 1)

        s = self.get_structure("Li3V2(PO4)3")
        p = DistinctPathFinder(s, "Li0+", max_path_length=4)
        paths = p.get_paths()

        self.assertEqual(len(paths), 4)
        p.write_all_paths("pathfindertest_LVPO.cif", nimages=10, idpp=True)

        for f in glob.glob("pathfindertest_*.cif"):
            os.remove(f)

    def test_max_path_length(self):
        s = Structure.from_file(get_path("LYPS.cif", dirname="pathfinder_files"))
        dp1 = DistinctPathFinder(s, "Li", perc_mode="1d")
        self.assertAlmostEqual(dp1.max_path_length, 4.11375354207, 7)
        dp2 = DistinctPathFinder(s, "Li", 5, perc_mode="1d")
        self.assertAlmostEqual(dp2.max_path_length, 5.0, 7)


class MigrationHopTest(PymatgenTest):
    def setUp(self):
        self.lifepo = self.get_structure("LiFePO4")
        m_graph = MigrationGraph.with_distance(self.lifepo, max_distance=4.0, migrating_specie="Li")
        gen = iter(m_graph.m_graph.graph.edges(data=True))
        u, v, d = next(gen)
        i_site = PeriodicSite("Li", coords=d["ipos"], lattice=self.lifepo.lattice)
        e_site = PeriodicSite("Li", coords=d["epos"], lattice=self.lifepo.lattice)
        a = SpacegroupAnalyzer(self.lifepo)
        symm_structure = a.get_symmetrized_structure()
        self.m_hop = MigrationHop(i_site, e_site, symm_structure)

    def test_msonable(self):
        hop_dict = self.m_hop.as_dict()
        assert type(hop_dict) == dict

    def test_get_start_end_structs_from_hop(self):
        dist_ref = self.m_hop.length
        base = self.lifepo.copy()
        base.remove_species(["Li"])
        start, end, b_sc = get_start_end_structures(
            self.m_hop.isite,
            self.m_hop.esite,
            base_struct=base,
            sc_mat=[[2, 1, 0], [-1, 1, 0], [0, 0, 1]],
            vac_mode=False,
        )
        start_site = next(filter(lambda x: x.species_string == "Li", start.sites))
        end_site = next(filter(lambda x: x.species_string == "Li", end.sites))
        self.assertAlmostEqual(start_site.distance(end_site), dist_ref, 3)

    def test_get_start_end_structs_from_hop_vac(self):
        dist_ref = self.m_hop.length
        start, end, b_sc = get_start_end_structures(
            self.m_hop.isite,
            self.m_hop.esite,
            base_struct=self.lifepo,
            sc_mat=[[2, 1, 0], [-1, 1, 0], [0, 0, 2]],
            vac_mode=True,
        )
        s1 = set()
        s2 = set()
        for itr_start, start_site in enumerate(start.sites):
            for itr_end, end_site in enumerate(end.sites):
                dist_ = start_site.distance(end_site)
                if dist_ < 1e-5:
                    s1.add(itr_start)
                    s2.add(itr_end)
        s1 = sorted(s1, reverse=True)
        s2 = sorted(s2, reverse=True)
        start.remove_sites(s1)
        end.remove_sites(s2)
        self.assertAlmostEqual(start.sites[0].distance(end.sites[0]), dist_ref, 3)

    def test_get_sc_structures(self):
        dist_ref = self.m_hop.length
        start, end, b_sc = self.m_hop.get_sc_structures(vac_mode=False)
        start_site = next(filter(lambda x: x.species_string == "Li", start.sites))
        end_site = next(filter(lambda x: x.species_string == "Li", end.sites))
        assert start.composition == end.composition == Composition("Li1 Fe24 P24 O96")
        assert b_sc.composition == Composition("Fe24 P24 O96")
        self.assertAlmostEqual(start_site.distance(end_site), dist_ref, 3)

    def test_get_sc_structures_vacmode(self):
        start, end, b_sc = self.m_hop.get_sc_structures(vac_mode=True)
        assert start.composition == end.composition == Composition("Li23 Fe24 P24 O96")
        assert b_sc.composition == Composition("Li24 Fe24 P24 O96")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Materials Virtual Lab.
# Distributed under the terms of the BSD License.

"""
Function to add edge data to MigrationGraph through 2 SC structures
"""

__author__ = "Haoming Li"
__copyright__ = "Copyright 2021, The Materials Project"
__maintainer__ = "Haoming Li"
__email__ = "HLi98@lbl.gov"
__date__ = "February 2, 2021"

import logging
from typing import Tuple

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core.structure import PeriodicSite, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph, MigrationHop
from pymatgen.analysis.diffusion.utils.parse_entries import get_matched_structure_mapping

logger = logging.getLogger(__name__)


def add_edge_data_from_sc(
    mg: MigrationGraph,
    i_sc: Structure,
    e_sc: Structure,
    data_array: list,
    key: str = "custom_key",
    use_host_sg: bool = True,
) -> None:
    """
    Add a data entry and key to edges within FullPathMapper object with the same hop_label.
    These hops are equivalent by symmetry to the 2 positions given in the supercell structures.

    Args:
        i_sc: Supercell structure containing working ion at initial position
        e_sc: Supercell structure containing working ion at ending position
        data_array: The data to be added to the edges
        key: Key of the edge attribute to be added
        use_host_sg: Flag t whether or not to use the host structure's spacegroup to initiate MigrationHop

    Returns:
        None
    """
    wi = list(mg.m_graph.graph.edges(data=True))[0][2]["hop"].isite.specie.name
    i_wi = [x for x in i_sc.sites if x.species_string == wi]
    e_wi = [x for x in e_sc.sites if x.species_string == wi]
    if len(i_wi) != 1 or len(e_wi) != 1:
        raise ValueError("The number of working ions in each supercell structure should be one")
    isite, esite = i_wi[0], e_wi[0]
    uhop_index, mh_from_sc = get_unique_hop(mg, i_sc, isite, esite, use_host_sg)
    add_dict = {key: data_array}
    mg.add_data_to_similar_edges(target_label=uhop_index, data=add_dict, m_hop=mh_from_sc)


def get_uc_pos(
    isite: PeriodicSite,
    esite: PeriodicSite,
    uc: Structure,
    sc: Structure,
    sm: StructureMatcher,
) -> Tuple[PeriodicSite, PeriodicSite, PeriodicSite]:
    """Take positions in the supercel and transform into the unitcell positions

    Args:
        isite: initial site in the SC
        esite: ending site in the SC
        uc: Unit Cell structre
        sc: Super Cell structure
        sm: StructureMatcher object with the working ion ignored

    Returns:
        The positions in the unit cell
    """
    mapping = get_matched_structure_mapping(base=uc, inserted=sc, sm=sm)
    if mapping is None:
        raise ValueError("Cannot obtain inverse mapping, consider lowering tolerances " "in StructureMatcher")
    sc_m, total_t = mapping
    sc_ipos = isite.frac_coords
    sc_ipos_t = sc_ipos - total_t
    uc_ipos = sc_ipos_t.dot(sc_m)
    image_trans = np.floor(uc_ipos)
    uc_ipos = uc_ipos - image_trans
    uc_ipos = _get_first_close_site(uc_ipos, uc)

    sc_epos = esite.frac_coords
    sc_epos_t = sc_epos - total_t
    uc_epos = sc_epos_t.dot(sc_m)
    uc_epos = uc_epos - image_trans
    uc_epos = _get_first_close_site(uc_epos, uc)

    sc_msite = PeriodicSite(
        esite.specie,
        (sc_ipos + sc_epos) / 2,
        esite.lattice,
    )
    sc_mpos = sc_msite.frac_coords

    sc_mpos_t = sc_mpos - total_t
    uc_mpos = sc_mpos_t.dot(sc_m)
    uc_mpos = uc_mpos - image_trans
    uc_mpos = _get_first_close_site(uc_mpos, uc)

    p0 = PeriodicSite(isite.specie, uc_ipos, uc.lattice)
    p1 = PeriodicSite(esite.specie, uc_mpos, uc.lattice)
    p2 = PeriodicSite(esite.specie, uc_epos, uc.lattice)
    return p0, p1, p2


def _get_first_close_site(frac_coord, structure, stol=0.1):
    for site in structure.sites:
        dist, image = structure.lattice.get_distance_and_image(frac_coord, site.frac_coords)
        if dist < stol:
            return np.add(site.frac_coords, image)
    return frac_coord


def mh_eq(mh1, mh2):
    """
    Allow for symmetric matching of MigrationPath objects with variable precession

    Args:
        mh1: MigrationHop object
        mh2: MigrationHop object

    Returns:
        Boolean True if they're equal, False if they are not
    """
    assert mh1.symm_structure == mh2.symm_structure
    return mh1 == mh2


def get_unique_hop(
    mg: MigrationGraph,
    sc: Structure,
    isite: PeriodicSite,
    esite: PeriodicSite,
    use_host_sg: bool = True,
) -> Tuple[int, MigrationHop]:
    """Get the unique hop label that correspond to two end positions in the SC

    Args:
        mg: Object containing the migration analysis
        sc: Structure of the supercell used for the NEB calculation
        isite: Initial position in the supercell
        esite: Final position in the supercell
        use_host_sg: Flag t whether or not to use the host structure's spacegroup to initiate MigrationHop

    Returns:
        The index of the unique hop, the MigrationHop object trasformed from the SC
    """
    sm = StructureMatcher(ignored_species=[list(mg.m_graph.graph.edges(data=True))[0][2]["hop"].isite.specie.name])
    uc_isite, uc_msite, uc_esite = get_uc_pos(isite, esite, mg.symm_structure, sc, sm)
    if use_host_sg:
        base_ss = SpacegroupAnalyzer(mg.host_structure, symprec=mg.symprec).get_symmetrized_structure()
        mh_from_sc = MigrationHop(
            uc_isite, uc_esite, symm_structure=mg.symm_structure, host_symm_struct=base_ss, symprec=mg.symprec
        )
    else:
        mh_from_sc = MigrationHop(uc_isite, uc_esite, symm_structure=mg.symm_structure, symprec=mg.symprec)
    result = []
    for k, v in mg.unique_hops.items():
        # tolerance may be changed here
        if mh_eq(v["hop"], mh_from_sc):
            result.append(k)

    if len(result) > 1:
        raise ValueError("Too many matches between UC and SC")
    if len(result) == 0:
        raise ValueError("No matches between UC and SC")

    assert mg.symm_structure.spacegroup.are_symmetrically_equivalent([uc_msite], [mh_from_sc.msite])
    return result[0], mh_from_sc
